//! d3-composer v0.6.7 - https://github.com/CSNW/d3-composer - @license: MIT
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('d3-array'), require('d3-selection'), require('d3-shape'), require('d3-axis')) :
  typeof define === 'function' && define.amd ? define(['exports', 'd3-array', 'd3-selection', 'd3-shape', 'd3-axis'], factory) :
  (global = global || self, factory(global.d3c = global.d3c || {}, global.d3, global.d3, global.d3, global.d3));
}(this, function (exports, d3Array, d3Selection, d3Shape, d3Axis) { 'use strict';

  function passthrough(d) {
    return d || [];
  }

  function array(value) {
    return typeof value === 'function'
      ? function(d, i, j) {
          return [value.call(this, d, i, j)];
        }
      : [value];
  }

  function fn(value) {
    return typeof value === 'function' ? value : function () { return value; };
  }

  function seriesExtent(data, values, value) {
    return data.reduce(
      function (ref, series, index) {
        var min = ref[0];
        var max = ref[1];

        var ref$1 = d3Array.extent(
          values(series, index, data),
          value
        );
        var series_min = ref$1[0];
        var series_max = ref$1[1];

        return [
          min != null && min < series_min ? min : series_min,
          max != null && max > series_max ? max : series_max
        ];
      },
      [undefined, undefined]
    );
  }

  var Size = d3Selection.local();
  var Area = d3Selection.local();
  var Series = d3Selection.local();

  var zero_margin = [0, 0, 0, 0];

  function toMargin(margin) {
    if (!margin) {
      return zero_margin;
    } else if (typeof margin === 'number') {
      return [margin, margin, margin, margin];
    } else if (!Array.isArray(margin)) {
      return [
        margin.top || 0,
        margin.right || 0,
        margin.bottom || 0,
        margin.left || 0
      ];
    } else {
      return margin;
    }
  }

  function toStyle(value) {
    if (!value) {
      return null;
    }
    if (typeof value === 'function') {
      return function() {
        var style = toStyle(value.apply(this, arguments));
        return typeof style === 'function' ? style.apply(this, arguments) : style;
      };
    }
    if (typeof value === 'string') {
      return value;
    }

    var ref = Object.keys(value).reduce(
      function (memo, key) {
        if (typeof value[key] !== 'function') {
          memo.fixed.push((key + ": " + (value[key]) + ";"));
        } else {
          memo.dynamic.push(function(d, i, j) {
            return (key + ": " + (value[key].call(this, d, i, j)) + ";");
          });
        }

        return memo;
      },
      { fixed: [], dynamic: [] }
    );
    var fixed = ref.fixed;
    var dynamic = ref.dynamic;

    if (!dynamic.length) {
      return fixed.join(' ');
    }

    return function(d, i, j) {
      var this$1 = this;

      return fixed.concat(dynamic.map(function (fn) { return fn.call(this$1, d, i, j); })).join(' ');
    };
  }

  var PARSE_ROW = /\"(.*?)\"\s*([\w\d]+)/g;
  var PARSE_MINMAX = /minmax\(([\w\d\s]+),([\w\d\s]+)\)/g;
  var PARSE_REPEAT = /repeat\((\d+),([^\)]+)\)/g;

  function parseTemplate(spec) {
    var formatted = spec
      .split('\n')
      .map(trim)
      .filter(Boolean)
      .join(' ');
    var ref = formatted.split('/', 2).map(trim);
    var areas_rows_spec = ref[0];
    var columns_spec = ref[1];

    if (!areas_rows_spec || !columns_spec) {
      throw new Error('Invalid grid template');
    }

    var areas_spec = '';
    var rows = [];
    var match;
    while ((match = PARSE_ROW.exec(areas_rows_spec))) {
      var _ = match[0];
      var area = match[1];
      var row = match[2];

      areas_spec += "\"" + area + "\"";
      rows.push(toLength(row));
    }

    var areas;
    if (!areas_spec) {
      rows = parseTracks(areas_rows_spec);
      areas = {};
    } else {
      areas = parseAreas(areas_spec);
    }

    var columns = parseTracks(columns_spec);

    return { areas: areas, rows: rows, columns: columns };
  }

  function parseTracks(spec) {
    var minmax_match;
    while ((minmax_match = PARSE_MINMAX.exec(spec))) {
      var result = minmax_match[0];
      var min = minmax_match[1];
      var max = minmax_match[2];

      // To handle potential nesting in repeat()
      // replace in spec with value that can be parsed by toLength
      var start = minmax_match.index;
      var end = start + result.length;
      var replacement = "MINMAX:" + (trim(min)) + "-" + (trim(max));
      spec = spec.substr(0, start) + replacement + spec.substr(end);
    }

    var repeat_match;
    while ((repeat_match = PARSE_REPEAT.exec(spec))) {
      var result$1 = repeat_match[0];
      var count = repeat_match[1];
      var value = repeat_match[2];
      var start$1 = repeat_match.index;
      var end$1 = start$1 + result$1.length;

      var replacement$1 = Array(parseInt(count, 10) + 1)
        .join((value + " "))
        .trim();

      spec = spec.substr(0, start$1) + replacement$1 + spec.substr(end$1);
    }

    var tracks = spec
      .split(' ')
      .filter(Boolean)
      .map(toLength);

    return tracks;
  }

  function parseAreas(spec) {
    var areas = {};

    spec
      .split('"')
      .filter(function (_, i) { return i % 2 !== 0; })
      .forEach(function (row_spec, row, rows) {
        var columns = row_spec.split(' ').filter(Boolean);

        columns.forEach(function (name, column) {
          if (name === '.') { return; }

          var next_row = rows[row + 1];
          var next_column = columns[column + 1];

          if (!areas[name]) {
            areas[name] = { top: row };
          }
          if (!('left' in areas[name])) {
            areas[name].left = column;
          }
          if (!next_row || next_row[column] !== name) {
            areas[name].bottom = row;
          }
          if (!next_column || next_column !== name) {
            areas[name].right = column;
          }
        });
      });

    return areas;
  }

  function toLength(value) {
    // units = 'fr' | 'percentage' | 'scalar'

    if (value.indexOf('MINMAX') === 0) {
      var ref = value.split(':');
      var _ = ref[0];
      var range = ref[1];
      var ref$1 = range.split('-');
      var min = ref$1[0];
      var max = ref$1[1];

      return [toLength(min), toLength(max)];
    }
    if (value === 'auto') {
      // For css grid, fr + auto would shrink auto to content
      // doesn't work well for this case
      // -> treat auto = 1fr
      return { value: 1, units: 'fr' };
    }
    if (endsWith(value, 'fr')) {
      return { value: Number(value.slice(0, -2)), units: 'fr' };
    }
    if (endsWith(value, '%')) {
      return { value: Number(value.slice(0, -1)) / 100, units: 'percentage' };
    }

    try {
      return { value: Number(value), units: 'scalar' };
    } catch (err) {
      throw new Error(("Unknown length value \"" + value + "\""));
    }
  }

  function trim(value) {
    return value.trim();
  }

  function endsWith(value, check) {
    if (!value || !check) { return false; }

    var index = value.indexOf(check);
    if (index < 0) { return false; }

    return index === value.length - check.length;
  }

  /*

  See complete spec at https://drafts.csswg.org/css-grid/#algo-track-sizing

  1. Initialize track sizes
    - base size, fixed = # or percentage
    - base size, flexible (fr or auto) = 0
    - growth limit, fixed = # or percentage
    - growth limit, flexible = base size
    - If growth limit is less than base size, increase growth limit to base size

  2. (skipped, no intrinsic track sizes in @d3-composer/grid)

  3. Maximize fixed tracks
    - S = grid size
    - B = combined base sizes
    - G = combined growth limits
    - A = available space
    - x = scaling factor
    
    S = Gx + A

    If G > S, A = 0 and x = S / G
    Otherwise, x = 1

  4. Expand flexible tracks
    - A = available space
    - F = sum of flex factors

    fr = A / F

    If calculated fr is less than track base size, re-calculate fr, treating track as inflexible

  5. (skipped, auto = 1fr in @d3-composer/grid)

  */

  function solve(lengths, size, gap, offset) {
    if ( gap === void 0 ) gap = 0;
    if ( offset === void 0 ) offset = 0;

    size -= gap * (lengths.length - 1);

    // 1. Initialize track sizes
    var base_total = 0;
    var growth_total = 0;

    lengths = lengths.map(function (length) {
      var flexible = false;
      var base = 0;
      var growth = 0;

      if (Array.isArray(length)) {
        if (length[0].units === 'percentage') {
          base = length[0].value * size;
        } else if (length[0].units === 'scalar') {
          base = length[0].value;
        } else {
          throw new Error(("Invalid units \"" + (length[0].units) + "\" for base value"));
        }

        if (length[1].units === 'scalar') {
          growth = length[1].value;
        } else if (length[1].units === 'percentage') {
          growth = length[1].value * size;
        } else {
          growth = length[1].value;
          flexible = true;
        }
      } else {
        if (length.units === 'scalar') {
          base = growth = length.value;
        } else if (length.units === 'percentage') {
          base = growth = length.value * size;
        } else {
          base = 0;
          growth = length.value;
          flexible = true;
        }
      }

      base_total += base;
      growth_total += flexible ? 0 : growth;

      return { base: base, growth: growth, flexible: flexible };
    });

    if (base_total > size) {
      throw new Error('More than 100% of available grid space was used');
    }

    // 2. (skipped)

    // 3. Maximize fixed tracks
    //
    // > If the free space is positive, distribute it equally to the base sizes of all tracks,
    // > freezing tracks as they reach their growth limits (and continuing to grow the unfrozen tracks as needed).
    //
    // Calculate a minimal step size so that the tracks increase equally

    if (growth_total > size) {
      var growth_size = base_total;
      var growing = lengths.filter(
        function (length) { return !length.flexible && length.growth > length.base; }
      );
      var available$1 = size - growth_size;
      var step_size = available$1 / growing.length;

      while (growth_size < size) {
        growing = growing.filter(function (length) {
          var remaining = length.growth - length.base;

          var step =
            available$1 < remaining && available$1 < step_size
              ? available$1
              : remaining < step_size
                ? remaining
                : step_size;

          length.base += step;
          growth_size += step;
          available$1 -= step;

          return length.growth > length.base;
        });
        step_size = available$1 / growing.length;
      }
    } else {
      // Growth is within free space, apply max growth to all
      lengths.forEach(function (length) {
        if (length.flexible) { return; }
        length.base = length.growth;
      });
    }

    // 4. Expand flexible tracks
    //
    // Calculate fr as available size / total factors
    // and apply unless growth is less than base,
    // in which case that track is made "inflexible"
    // and fr is re-calculated

    var available = growth_total > size ? 0 : size - growth_total;
    if (available) {
      var flexible = lengths.filter(function (ref) {
        var flexible = ref.flexible;

        return flexible;
      });
      var solved = flexible.length === 0;

      var loop = function () {
        var factor_total = flexible.reduce(
          function (memo, ref) {
            var growth = ref.growth;

            return memo + growth;
        },
          0
        );
        var fr = available / factor_total;

        var valid = flexible.filter(function (length) {
          var base = length.base;
          var growth = length.growth;
          var value = growth * fr;

          if (value < base) {
            length.flexible = false;
            return false;
          } else {
            // If lengths are made inflexible,
            // fr is guaranteed to go up and base will increase
            // -> safe to set base here

            length.base = value;
            return true;
          }
        });

        if (valid.length === flexible.length) {
          solved = true;
        } else {
          flexible = valid;
        }
      };

      while (!solved) loop();
    }

    // 5. (skipped)

    // Finally, extract [left, right] values for each track
    return lengths.reduce(function (memo, length, index) {
      var last = memo[index - 1];

      var left = last ? last[1] + gap : offset;
      var right = left + length.base;

      memo.push([left, right]);
      return memo;
    }, []);
  }

  var cache = new Map();

  function template(spec, options) {
    var assign;

    var width = options.width;
    var height = options.height;
    var gap = options.gap;
    var row_gap = options.row_gap;
    var column_gap = options.column_gap;
    var padding = options.padding;
    padding = toMargin(padding);
    width -= padding[1] + padding[3];
    height -= padding[0] + padding[2];
    if (gap) { row_gap = column_gap = gap; }

    // spec: string = template
    // spec: object = { rows, columns, [areas] } templates
    var rows, columns, areas;
    if (typeof spec === 'object') {
      rows = parseTracks(spec.rows);
      columns = parseTracks(spec.columns);
      areas = spec.areas ? parseAreas(spec.areas) : {};
    } else {
      var parsed = cache.has(spec) ? cache.get(spec) : parseTemplate(spec);
      cache.set(spec, parsed);
      ((assign = parsed, rows = assign.rows, columns = assign.columns, areas = assign.areas));
    }

    var x = solve(columns, width, column_gap, padding[3]);
    var y = solve(rows, height, row_gap, padding[0]);

    return layout(rows, columns, areas, x, y);
  }

  function layout(rows, columns, areas, x, y) {
    var result = { rows: y, columns: x, items: [], areas: {} };

    for (var i = 0; i < rows.length; i++) {
      for (var j = 0; j < columns.length; j++) {
        var top = y[i][0];
        var right = x[j][1];
        var bottom = y[i][1];
        var left = x[j][0];

        result.items.push({
          x: left,
          y: top,
          width: right - left,
          height: bottom - top,
          top: top,
          right: right,
          bottom: bottom,
          left: left
        });
      }
    }

    Object.keys(areas).forEach(function (name) {
      var area = areas[name];

      var top = y[area.top][0];
      var right = x[area.right][1];
      var bottom = y[area.bottom][1];
      var left = x[area.left][0];

      result.areas[name] = {
        x: left,
        y: top,
        width: right - left,
        height: bottom - top,
        top: top,
        right: right,
        bottom: bottom,
        left: left
      };
    });

    return result;
  }

  function applyMargin(area, margin) {
    margin = toMargin(margin);

    var top = area.top;
    var right = area.right;
    var bottom = area.bottom;
    var left = area.left;
    top += margin[0];
    right -= margin[1];
    bottom -= margin[2];
    left += margin[3];

    return {
      x: left,
      y: top,
      width: right - left,
      height: bottom - top,
      top: top,
      right: right,
      bottom: bottom,
      left: left
    };
  }

  var size_id = Size.toString();

  /**
   * Determine node size for given selection
   *
   * 1. Check viewBox (if svg)
   * 2. Check attributes (if svg)
   * 3. Check Size local
   * 4. Check bbox
   *
   * @param {d3.selection} selection
   */
  function size(selection) {
    var node = selection.node();

    if (node.tagName === 'svg') {
      var viewBox = node.getAttribute('viewBox');
      if (viewBox) {
        var ref = viewBox.split(' ');
        var _1 = ref[0];
        var _2 = ref[1];
        var width = ref[2];
        var height = ref[3];
        return { width: Number(width), height: Number(height) };
      }

      var width$1 = node.getAttribute('width');
      var height$1 = node.getAttribute('height');

      if (width$1 != null && height$1 != null) {
        return { width: Number(width$1), height: Number(height$1) };
      }
    }

    if (size_id in node) {
      return node[size_id];
    }

    try {
      var bbox = node.getBBox();
      return { width: bbox.width, height: bbox.height };
    } catch (err) {
      return { width: NaN, height: NaN };
    }
  }

  function childNodes() {
    return this.childNodes;
  }

  function interpolatePath(selection, path, interpolate) {
    if (interpolate) {
      selection.attrTween('d', function(d, i, j) {
        var next = typeof path === 'function' ? path.call(this, d, i, j) : path;
        var previous = this.getAttribute('d');

        return interpolate(previous, next);
      });
    } else {
      selection.attr('d', path);
    }
  }

  function translateXY(_x, _y) {
    return function(d, i, j) {
      var x = _x.call(this, d, i, j);
      var y = _y.call(this, d, i, j);

      return ("translate(" + x + ", " + y + ")");
    };
  }

  var assign =
    Object.assign ||
    function(target) {
      var objects = [], len = arguments.length - 1;
      while ( len-- > 0 ) objects[ len ] = arguments[ len + 1 ];

      objects.forEach(function (object) {
        Object.keys(object).forEach(function (key) {
          target[key] = object[key];
        });
      });

      return target;
    };

  function byIndex(_d, i) {
    return i;
  }

  function chart(selection, options) {
    if ( options === void 0 ) options = {};

    var width = options.width; if ( width === void 0 ) width = 400;
    var height = options.height; if ( height === void 0 ) height = 300;
    var responsive = options.responsive; if ( responsive === void 0 ) responsive = false;

    var svg;
    if (selection.node().tagName === 'svg') {
      svg = selection;
    } else {
      var ratio = (height / width) * 100;
      var style = responsive
        ? ("position: relative; width: 100%; height: 0; padding-top: " + ratio + "%;")
        : null;

      selection.attr('data-container', '').attr('style', style);
      svg = svgLayer(selection, { responsive: responsive });
    }

    return svg
      .attr('viewBox', responsive ? ("0 0 " + width + " " + height) : null)
      .attr('preserveAspectRatio', responsive ? 'xMidYMid meet' : null)
      .attr('width', responsive ? null : width)
      .attr('height', responsive ? null : height);
  }

  function svgLayer(selection, options) {
    var responsive = options.responsive;

    var style = responsive ? 'position: absolute; top: 0; left: 0;' : null;

    var svg = selection
      .selectAll(childNodes)
      .filter('svg')
      .data([null]);

    return svg
      .enter()
      .append('svg')
      .merge(svg)
      .attr('style', style);
  }

  function layer(selection, id, options) {
    if ( options === void 0 ) options = {};

    var element = options.element; if ( element === void 0 ) element = 'g';
    var instance = selection
      .selectAll(childNodes)
      .filter(("[data-layer=\"" + id + "\"]"))
      .data([null]);

    // Forward size from selection to layer
    var size = selection.property(Size);

    return instance
      .enter()
      .append(element)
      .attr('data-layer', id)
      .merge(instance)
      .property(Size, size);
  }

  function layer$1(selection, id, area, options) {
    var margin = options.margin;
    var style = options.style;
    var className = options.class;
    area = applyMargin(area, margin);

    return layer(selection, id)
      .property(Area, area)
      .property(Size, area)
      .attr('style', toStyle(style))
      .attr('class', className)
      .attr('transform', ("translate(" + (area.x) + ", " + (area.y) + ")"));
  }

  function layout$1(selection, grid, callback) {
    var index = 0;
    var indices = {};
    var added = [];

    function layers(id, options) {
      if ( options === void 0 ) options = {};

      if (id && typeof id === 'object') {
        options = id;
        id = undefined;
      }

      var item = index++;
      if (index >= grid.items.length) { index = 0; }

      var name = "__item_" + item;
      if (!indices[name]) { indices[name] = 0; }
      var i = indices[name]++;

      id = id || ("item-" + item + (i > 0 ? ("-" + i) : ''));
      added.push(id);

      return layer$1(selection, id, grid.items[item], options).attr(
        'data-item',
        item
      );
    }

    Object.keys(grid.areas).forEach(function (name) {
      var area = grid.areas[name];
      indices[name] = 0;

      layers[name] = function (id, options) {
        if ( options === void 0 ) options = {};

        if (id && typeof id === 'object') {
          options = id;
          id = undefined;
        }
        var i = indices[name]++;

        id = id || ("area-" + name + (i > 0 ? ("-" + i) : ''));
        added.push(id);

        return layer$1(selection, id, area, options).attr('data-area', name);
      };
    });

    // TODO This isn't ideomatic d3, but not sure the best way to deal with dynamic layers
    callback(layers);

    // Remove stale layers
    selection
      .selectAll(childNodes)
      .filter('[data-layer]')
      .data(added, function(d) {
        return (
          d ||
          (this.dataset ? this.dataset.layer : this.getAttribute('data-layer'))
        );
      })
      .order()
      .exit()
      .remove();

    return selection;
  }

  var seriesKey = function (d, i) { return (d && d.key != null ? d.key : i); };

  function series(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var key = options.key; if ( key === void 0 ) key = seriesKey;
    var style = options.style;
    var className = options.class;

    var layers = selection
      .selectAll(childNodes)
      .filter('[data-series]')
      .data(data, key);

    layers.exit().remove();

    return layers
      .enter()
      .append('g')
      .attr('data-series', key)
      .property(Series, function (d) { return d; })
      .merge(layers)
      .attr('class', className)
      .attr('style', toStyle(style))
      .order();
  }

  function stack(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var key = options.key; if ( key === void 0 ) key = byIndex;
    var direction = options.direction; if ( direction === void 0 ) direction = 'vertical';
    var _size = options.size;
    var className = options.class;
    var style = options.style;
    var size$1 = fn(_size);
    var ref = size(selection);
    var width = ref.width;
    var height = ref.height;

    var items = selection.selectAll('g').data(data, key);

    items.exit().remove();

    var sizes = [];
    var offset = 0;

    return items
      .enter()
      .append('g')
      .merge(items)
      .property(Size, function(d, i, j) {
        sizes[i] = size$1.call(this, d, i, j) || 0;

        return direction === 'vertical'
          ? { width: width, height: sizes[i] }
          : { width: sizes[i], height: height };
      })
      .attr('transform', function (_d, i) {
        var x = direction === 'vertical' ? 0 : offset;
        var y = direction === 'vertical' ? offset : 0;
        offset += sizes[i];

        return ("translate(" + x + ", " + y + ")");
      })
      .attr('style', toStyle(style))
      .attr('class', className);
  }

  function vstack(selection, options) {
    return stack(selection, assign({ direction: 'vertical' }, options));
  }

  function hstack(selection, options) {
    return stack(selection, assign({ direction: 'horizontal' }, options));
  }

  function line(selection, options) {
    if ( options === void 0 ) options = {};

    var data = options.data; if ( data === void 0 ) data = passthrough;
    var x = options.x;
    var y = options.y;
    var key = options.key;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;
    var curve = options.curve;
    var interpolate = options.interpolate;
    var defined = options.defined;

    var line = d3Shape.line()
      .x(x)
      .y(y);
    if (curve) { line.curve(curve); }
    if (defined) { line.defined(defined); }

    var path = selection.selectAll('path').data(array(data), key);

    path.exit().remove();

    path
      .enter()
      .append('path')
      .attr('d', line)
      .merge(path)
      .attr('fill', 'none')
      .attr('stroke', 'currentColor')
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .call(interpolatePath, line, interpolate);

    return selection;
  }

  function bars(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var x = options.x;
    var x0 = options.x0;
    var x1 = options.x1;
    var y = options.y;
    var y0 = options.y0;
    var y1 = options.y1;
    var width = options.width;
    var height = options.height;
    var key = options.key;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;

    x = x != null ? x : x0;
    y = y != null ? y : y1;

    width =
      width ||
      function(d, i, j) {
        return Math.abs(fn(x1).call(this, d, i, j) - fn(x).call(this, d, i, j));
      };
    height =
      height ||
      function(d, i, j) {
        return Math.abs(fn(y).call(this, d, i, j) - fn(y0).call(this, d, i, j));
      };

    var bars = selection.selectAll('rect').data(data, key);

    bars.exit().remove();
    bars
      .enter()
      .append('rect')
      .attr('x', x)
      .attr('y', y)
      .attr('height', height)
      .attr('width', width)
      .merge(bars)
      .attr('fill', 'currentColor')
      .attr('stroke', 'none')
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .attr('x', x)
      .attr('y', y)
      .attr('height', height)
      .attr('width', width);

    return selection;
  }

  function scatter(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var x = options.x;
    var y = options.y;
    var key = options.key;
    var path = options.path;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;
    var interpolate = options.interpolate;
    var translate = translateXY(x, y);

    var paths = selection.selectAll('path').data(data, key);

    paths.exit().remove();

    paths
      .enter()
      .append('path')
      .attr('transform', translate)
      .merge(paths)
      .attr('fill', 'currentColor')
      .attr('stroke', 'none')
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .attr('transform', translate)
      .call(interpolatePath, path, interpolate);

    return selection;
  }

  function area(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var x = options.x;
    var x0 = options.x0;
    var x1 = options.x1;
    var y0 = options.y0;
    var y1 = options.y1;
    var key = options.key;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;
    var curve = options.curve;
    var interpolate = options.interpolate;
    var defined = options.defined;

    var area = d3Shape.area()
      .y0(y0)
      .y1(y1);
    if (x) { area.x(x); }
    if (x0 && x1) { area.x0(x0).x1(x1); }
    if (curve) { area.curve(curve); }
    if (defined) { area.defined(defined); }

    var path = selection.selectAll('path').data(array(data), key);

    path
      .exit()
      .transition(transition)
      .remove();
    path
      .enter()
      .append('path')
      .attr('d', area)
      .merge(path)
      .attr('fill', 'currentColor')
      .attr('stroke', 'none')
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .call(interpolatePath, area, interpolate);

    return selection;
  }

  function labels(selection, options) {
    var data = options.data; if ( data === void 0 ) data = passthrough;
    var x = options.x;
    var y = options.y;
    var key = options.key;
    var text = options.text;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;
    var transform = options.transform;
    var anchor = options.anchor; if ( anchor === void 0 ) anchor = 'start';
    var baseline = options.baseline; if ( baseline === void 0 ) baseline = 'hanging';
    var translate = translateXY(x, y);

    var groups = selection.selectAll('g').data(data, key);

    groups.exit().remove();

    var entering = groups
      .enter()
      .append('g')
      .attr('transform', translate);

    entering
      .merge(groups)
      .transition(transition)
      .attr('transform', translate);
    entering
      .append('text')
      .merge(groups.select('text'))
      .attr('text-anchor', anchor)
      .attr('dominant-baseline', baseline)
      .attr('style', toStyle(style))
      .attr('class', className)
      .attr('transform', transform || null)
      .text(text);

    return selection;
  }

  function axisTop(selection, options) {
    var scale = options.scale;
    var xScale = options.xScale;
    var style = options.style;
    var domainStyle = options.domainStyle;
    var tickStyle = options.tickStyle;
    var labelStyle = options.labelStyle;
    var className = options.class;
    var transition = options.transition;
    var axis = prepare(d3Axis.axisTop(scale || xScale), options);
    var ref = size(selection);
    var height = ref.height;

    layer(selection, 'axis-transform')
      .attr('transform', ("translate(0, " + (height || 0) + ")"))
      .attr('class', className)
      .transition(transition)
      .call(axis)
      .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

    return selection;
  }

  function axisRight(selection, options) {
    var scale = options.scale;
    var yScale = options.yScale;
    var style = options.style;
    var domainStyle = options.domainStyle;
    var tickStyle = options.tickStyle;
    var labelStyle = options.labelStyle;
    var className = options.class;
    var transition = options.transition;
    var axis = prepare(d3Axis.axisRight(scale || yScale), options);

    selection
      .attr('class', className)
      .transition(transition)
      .call(axis)
      .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

    return selection;
  }

  function axisBottom(selection, options) {
    var scale = options.scale;
    var xScale = options.xScale;
    var style = options.style;
    var domainStyle = options.domainStyle;
    var tickStyle = options.tickStyle;
    var labelStyle = options.labelStyle;
    var className = options.class;
    var transition = options.transition;
    var axis = prepare(d3Axis.axisBottom(scale || xScale), options);

    selection
      .attr('class', className)
      .transition(transition)
      .call(axis)
      .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

    return selection;
  }

  function axisLeft(selection, options) {
    var scale = options.scale;
    var yScale = options.yScale;
    var style = options.style;
    var domainStyle = options.domainStyle;
    var tickStyle = options.tickStyle;
    var labelStyle = options.labelStyle;
    var className = options.class;
    var transition = options.transition;
    var axis = prepare(d3Axis.axisLeft(scale || yScale), options);
    var ref = size(selection);
    var width = ref.width;

    layer(selection, 'axis-transform')
      .attr('transform', ("translate(" + (width || 0) + ", 0)"))
      .attr('class', className)
      .transition(transition)
      .call(axis)
      .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

    return selection;
  }

  function prepare(axis, options) {
    var ticks = options.ticks;
    var tickArguments = options.tickArguments;
    var tickValues = options.tickValues;
    var tickFormat = options.tickFormat;
    var tickSize = options.tickSize;
    var tickSizeInner = options.tickSizeInner;
    var tickSizeOuter = options.tickSizeOuter;
    var tickPadding = options.tickPadding;

    if (ticks) { axis.ticks(ticks); }
    if (tickArguments) { axis.tickArguments(tickArguments); }
    if (tickValues) { axis.tickValues(tickValues); }
    if (tickFormat) { axis.tickFormat(tickFormat); }
    if (tickSize) { axis.tickSize(tickSize); }
    if (tickSizeInner) { axis.tickSizeInner(tickSizeInner); }
    if (tickSizeOuter) { axis.tickSizeOuter(tickSizeOuter); }
    if (tickPadding) { axis.tickPadding(tickPadding); }

    return function(selection) {
      axis.call(this, selection);

      // Reset hard-coded text values
      selection.attr('font-size', null).attr('font-family', null);
    };
  }

  function styleAxis(selection, style, domainStyle, tickStyle, labelStyle) {
    // Get underlying selection from transition
    selection = selection.selection ? selection.selection() : selection;

    selection.attr('style', toStyle(style));
    selection.select('.domain').attr('style', toStyle(domainStyle));
    selection.selectAll('.tick line').attr('style', toStyle(tickStyle));
    selection.selectAll('.tick text').attr('style', toStyle(labelStyle));
  }

  function text(selection, options) {
    // anchor = origin-x = 'start' | 'middle' | 'end'
    //   (https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/text-anchor)
    // baseline = origin-y = 'hanging' | 'middle' | 'baseline'
    //   (https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/dominant-baseline)
    // justify = container-x = 'start' | 'center' | 'end'
    //   (https://css-tricks.com/snippets/css/complete-guide-grid/#article-header-id-30)
    // align = container-y = 'start' | 'center' | 'end'
    //   (https://css-tricks.com/snippets/css/complete-guide-grid/#article-header-id-31)

    var text = options.text;
    var anchor = options.anchor;
    var baseline = options.baseline;
    var justify = options.justify; if ( justify === void 0 ) justify = 'start';
    var align = options.align; if ( align === void 0 ) align = 'start';
    var className = options.class;
    var style = options.style;
    var rotation = options.rotation;
    var transform = options.transform;

    anchor =
      anchor != null
        ? anchor
        : justify === 'center'
          ? 'middle'
          : justify === 'end'
            ? 'end'
            : 'start';

    baseline =
      baseline != null
        ? baseline
        : align === 'center'
          ? 'middle'
          : align === 'end'
            ? 'baseline'
            : 'hanging';

    if (!transform && (justify !== 'start' || align !== 'start' || rotation)) {
      var ref = size(selection);
      var width = ref.width;
      var height = ref.height;
      var x = 0;
      var y = 0;

      if (justify === 'center') {
        x = width / 2;
      } else if (justify === 'end') {
        x = width;
      }

      if (align === 'center') {
        y = height / 2;
      } else if (align === 'end') {
        y = height;
      }

      transform = "translate(" + x + ", " + y + ") rotate(" + (rotation || 0) + ")";
    }

    var instance = selection.selectAll('text').data([null]);
    instance
      .enter()
      .append('text')
      .merge(instance)
      .attr('dominant-baseline', baseline)
      .attr('text-anchor', anchor)
      .attr('transform', transform)
      .attr('class', className)
      .attr('style', toStyle(style))
      .text(text);

    return selection;
  }

  function legend(selection, options) {
    var path = options.path;
    var size = options.size;
    var width = options.width;
    var height = options.height;
    var text = options.text; if ( text === void 0 ) text = function (d) { return d; };
    var data = options.data;
    var groupStyle = options.groupStyle;
    var groupClass = options.groupClass;
    var pathStyle = options.pathStyle;
    var labelStyle = options.labelStyle;

    // TODO Accept size, width, or height as functions
    if (!height && !size) {
      size = 50;
    }
    if (!height) {
      height = 3 * Math.sqrt(size);
    }
    if (!width) {
      width = height;
    }
    if (!size) {
      size = (Math.min(width, height) * Math.min(width, height)) / 3;
    }

    if (!path) {
      path = d3Shape.symbol()
        .size(size)
        .type(d3Shape.symbolCircle);
    }

    var layers = selection
      .selectAll('g')
      .data(data, function (d, i) { return (d && d.key != null ? d.key : i); });

    layers.exit().remove();

    var entering = layers.enter().append('g');

    entering
      .merge(layers)
      .attr('transform', function (_, i) { return ("translate(0, " + (i * height) + ")"); })
      .attr('style', toStyle(groupStyle))
      .attr('class', groupClass);
    entering
      .append('path')
      .merge(layers.select('path'))
      .attr('transform', ("translate(" + (width / 2) + ", " + (height / 2) + ")"))
      .attr('fill', 'currentColor')
      .attr('stroke', 'none')
      .attr('style', toStyle(pathStyle))
      .attr('d', path);
    entering
      .append('text')
      .merge(layers.select('text'))
      .attr('dominant-baseline', 'central')
      .attr('transform', ("translate(" + width + ", " + (height / 2) + ")"))
      .attr('fill', 'currentColor')
      .attr('stroke', 'none')
      .attr('style', toStyle(labelStyle))
      .text(text);

    return selection;
  }

  var index = function (_, i) { return i; };

  function gridlines(selection, options) {
    var xGrid = options.xGrid; if ( xGrid === void 0 ) xGrid = true;
    var yGrid = options.yGrid; if ( yGrid === void 0 ) yGrid = true;
    var xScale = options.xScale;
    var yScale = options.yScale;
    var xTicks = options.xTicks;
    var xTickArguments = options.xTickArguments;
    var xTickValues = options.xTickValues;
    var yTicks = options.yTicks;
    var yTickArguments = options.yTickArguments;
    var yTickValues = options.yTickValues;
    var style = options.style;
    var className = options.class;
    var transition = options.transition;

    var ref = xScale.range();
    var x1 = ref[0];
    var x2 = ref[1];
    var ref$1 = yScale.range();
    var y1 = ref$1[0];
    var y2 = ref$1[1];

    var translateX = function (d) { return ("translate(" + (xScale(d) + 0.5) + ", 0)"); };
    var translateY = function (d) { return ("translate(0, " + (yScale(d) + 0.5) + ")"); };

    var x_data = !xGrid
      ? []
      : xTickValues ||
        xScale.ticks.apply(
          xScale.ticks,
          xTicks ? [xTicks] : xTickArguments || []
        );
    var y_data = !yGrid
      ? []
      : yTickValues ||
        yScale.ticks.apply(
          yScale.ticks,
          yTicks ? [yTicks] : yTickArguments || []
        );

    var x_lines = layer(selection, 'x')
      .selectAll('line')
      .data(x_data, index);

    x_lines.exit().remove();
    x_lines
      .enter()
      .append('line')
      .attr('transform', translateX)
      .merge(x_lines)
      .attr('x1', 0)
      .attr('x2', 0)
      .attr('y1', y1)
      .attr('y2', y2)
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .attr('transform', translateX);

    var y_lines = layer(selection, 'y')
      .selectAll('line')
      .data(y_data, index);

    y_lines.exit().remove();
    y_lines
      .enter()
      .append('line')
      .attr('transform', translateY)
      .merge(y_lines)
      .attr('x1', x1)
      .attr('x2', x2)
      .attr('y1', 0)
      .attr('y2', 0)
      .attr('style', toStyle(style))
      .attr('class', className)
      .transition(transition)
      .attr('transform', translateY);

    return selection;
  }

  var symbolLine = {
    draw: function draw(context, size) {
      var w = Math.sqrt(size) * 2;
      var h = size / w;
      var x = -w / 2;
      var y = -h / 2;

      context.rect(x, y, w, h);
    }
  };

  exports.Area = Area;
  exports.Series = Series;
  exports.Size = Size;
  exports.applyMargin = applyMargin;
  exports.area = area;
  exports.array = array;
  exports.axisBottom = axisBottom;
  exports.axisLeft = axisLeft;
  exports.axisRight = axisRight;
  exports.axisTop = axisTop;
  exports.bars = bars;
  exports.chart = chart;
  exports.gridlines = gridlines;
  exports.hstack = hstack;
  exports.interpolatePath = interpolatePath;
  exports.labels = labels;
  exports.layer = layer;
  exports.layout = layout$1;
  exports.legend = legend;
  exports.line = line;
  exports.passthrough = passthrough;
  exports.scatter = scatter;
  exports.series = series;
  exports.seriesExtent = seriesExtent;
  exports.size = size;
  exports.stack = stack;
  exports.symbolLine = symbolLine;
  exports.template = template;
  exports.text = text;
  exports.toMargin = toMargin;
  exports.toStyle = toStyle;
  exports.translateXY = translateXY;
  exports.vstack = vstack;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
